"""
Python script to use with pam_python.so to authenticate LDAP users with their mail access tokens
"""
# pylint: disable=c-extension-no-member

import argparse
import configparser
import datetime
import json
import logging
import sys

import dateutil.parser
import ldap
import PAM
from argon2 import PasswordHasher
from argon2.exceptions import VerifyMismatchError
from ldap.filter import escape_filter_chars

log = logging.getLogger(__name__)


def handle_args(argv):
    """Handle arguments provided from PAM configuration"""
    parser = argparse.ArgumentParser()
    parser.add_argument("config", help="Config file path", default="/etc/ee_pam_mailaccess.ini")
    parser.add_argument("-d", "--debug", action="store_true", help="Enable debug mode")
    parser.add_argument("-v", "--verbose", action="store_true", help="Enable verbose mode")
    args = parser.parse_args(argv[1:])

    # Read configuration file
    config = configparser.ConfigParser()
    if args.config:
        config.read(args.config)

    # Initialize logging
    log_config = config["log"] if "log" in config else {}
    logging.basicConfig(
        level=(
            logging.DEBUG
            if args.debug
            else logging.INFO
            if args.verbose
            else log_config.get("level", "WARNING")
        ),
        format=log_config.get("format", "%(asctime)s - %(levelname)s - %(message)s"),
        handlers=[
            logging.FileHandler(log_config["path"])
            if log_config.get("path")
            else logging.StreamHandler(sys.stderr)
        ],
    )

    return config


# pylint: disable=too-many-locals, too-many-statements
def check_credentials(config, username, password):
    """
    Check if user specified credentials match with an existing user and one of their valid
    mail access tokens
    """

    # Check LDAP configuration
    ldap_config = config["ldap"] if "ldap" in config else {}
    ldap_uri = ldap_config.get("uri", "ldap://localhost")
    ldap_bind_dn = ldap_config.get("bind_dn")
    ldap_bind_password = ldap_config.get("bind_password")
    user_search_base_dn = ldap_config.get("user_search_base_dn")
    user_search_filter = ldap_config.get("user_search_filter", "(uid=[username])")
    user_search_scope = ldap_config.get("user_search_scope", "sub")
    try:
        getattr(ldap, f"SCOPE_{user_search_scope.upper()}")
    except AttributeError:
        log.error("Invalid LDAP user search scope configured (%s)", user_search_scope)
    user_mail_access_token_attr = ldap_config.get(
        "user_mail_access_token_attr", "eeMailAccessToken"
    )

    assert (
        user_search_base_dn
    ), "Invalid LDAP configuration: You must at least configure user_search_base_dn."

    log.debug("%s: connect to LDAP host %s", username, ldap_uri)
    ldap_client = ldap.initialize(ldap_uri)
    if ldap_bind_dn and ldap_bind_password:
        log.debug("%s: auth on %s as %s", username, ldap_uri, ldap_bind_dn)
        ldap_client.simple_bind_s(ldap_bind_dn, ldap_bind_password)

    search_filter = user_search_filter.replace("[username]", escape_filter_chars(username))
    log.debug(
        "%s: search user in LDAP directory on base DN '%s' with filter '%s'",
        username,
        user_search_base_dn,
        search_filter,
    )
    result = ldap_client.search_s(
        user_search_base_dn,
        user_search_scope,
        search_filter,
        [user_mail_access_token_attr],
    )

    if not result:
        log.info("%s: not found in LDAP directory", username)
        return False

    user_dn, user_attrs = result[0]
    log.info("%s: found in LDAP directory as '%s'", username, user_dn)
    token_values = user_attrs.get(user_mail_access_token_attr, [])
    if not token_values:
        log.info("%s: no mail access token in LDAP", username)
        return False

    password_hasher = PasswordHasher()
    for token_value in token_values:
        try:
            token_data = json.loads(token_value.decode("utf-8"))
        except ValueError:
            log.exception(
                "Invalid JSON data found in %s attribute of user %s (%s): `%s`",
                user_mail_access_token_attr,
                username,
                user_dn,
                token_value,
            )
            continue
        if "hash" not in token_data:
            log.warning(
                "No token hash found in %s attribute of user %s (%s): `%s`",
                user_mail_access_token_attr,
                username,
                user_dn,
                token_value,
            )
            continue
        token_hash = token_data["hash"]
        label = token_data.get("label", "[no label]")

        # Check expiration
        expiration = token_data.get("expiration")
        if expiration:
            expiration_date = dateutil.parser.parse(expiration, dayfirst=False).replace(tzinfo=None)
            if expiration_date < datetime.datetime.now():
                log.info(
                    "%s: ignore expired mail access token %s (since %s)",
                    username,
                    label,
                    expiration_date,
                )
                continue
            log.debug(
                "%s: mail access token '%s' still valid until %s",
                username,
                label,
                expiration_date,
            )
        else:
            log.debug("%s: mail access token '%s' will never expire", username, label)

        # Check token hash
        try:
            if password_hasher.verify(token_hash, password):
                log.info(
                    "%s: authentication successful with mail access token '%s'",
                    username,
                    label,
                )
                return True
        except VerifyMismatchError:
            log.debug(
                "%s: mail access token '%s' does not match with provided password",
                username,
                label,
            )
            continue

    log.info("%s: no match found with user's mail access tokens", username)
    return False


def pam_sm_authenticate(pamh, flags, argv):  # pylint: disable=unused-argument
    """
    The service module's implementation of PAM's :manpage:`pam_authenticate(3)`
    interface.

    This function performs the task of authenticating the user.
    """
    config = handle_args(argv)

    try:
        # Retrieve username
        username = pamh.get_user(None)
        if not username:
            return PAM.PAM_USER_UNKNOWN

        # Retrieve (and ask if need) password
        password = pamh.authtok
        if not password:
            password = pamh.conversation(PAM.PAM_PROMPT_ECHO_OFF, "Mail access token: ").resp
            if not password:
                return PAM.PAM_AUTH_ERR

        # Check authentication
        return (
            PAM.PAM_SUCCESS if check_credentials(config, username, password) else PAM.PAM_AUTH_ERR
        )
    except ldap.INVALID_CREDENTIALS:  # pylint: disable=no-member
        return PAM.PAM_USER_UNKNOWN
    except pamh.exception as err:
        return err.pam_result


def pam_sm_setcred(pamh, flags, argv):  # pylint: disable=unused-argument
    """
    The service module's implementation of PAM's :manpage:`pam_setcred(3)` interface.

    This function performs the task of altering the credentials of the user with respect to the
    corresponding authorization scheme. Generally, an authentication module may have access to more
    information about a user than their authentication token. This function is used to make such
    information available to the application. It should only be called after the user has been
    authenticated but before a session has been established.
    """
    return PAM.PAM_CRED_UNAVAIL


def pam_sm_acct_mgmt(pamh, flags, argv):  # pylint: disable=unused-argument
    """
    The service module's implementation of PAM's :manpage:`pam_acct_mgmt(3)` interface.

    This function performs the task of establishing whether the user is permitted to gain access
    at this time. It should be understood that the user has previously been validated by an
    authentication module. This function checks for other things. Such things might be: the time of
    day or the date, the terminal line, remote hostname, etc. This function may also determine
    things like the expiration on passwords, and respond that the user change it before continuing.
    """
    return PAM.PAM_SUCCESS


def pam_sm_open_session(pamh, flags, argv):  # pylint: disable=unused-argument
    """
    The service module's implementation of PAM's :manpage:`pam_open_session(3)`
    interface.

    This function is called to commence a session.
    """
    return PAM.PAM_SUCCESS


def pam_sm_close_session(pamh, flags, argv):  # pylint: disable=unused-argument
    """
    The service module's implementation of PAM's :manpage:`pam_close_session(3)`
    interface.

    This function is called to terminate a session.
    """
    return PAM.PAM_SUCCESS


def pam_sm_chauthtok(pamh, flags, argv):  # pylint: disable=unused-argument
    """
    The service module's implementation of PAM's :manpage:`pam_chauthtok(3)` interface.

    This function is used to (re-)set the authentication token of the user.
    """
    return PAM.PAM_SUCCESS
